# DATA

# data = [[80, 40, 60, 20, 0,0,0,0,0,0,0,0,200]]
MODE_ANIMATION = "max"

barHeight = 0.001
barLabels = ["Pulp Fiction", "Jacky Brown", "The Hateful Eight", "Movie 4", "Movie 5", "Movie 6", "Movie very Long Placeholder 7", "Movie 8", "Movie 9", "Movie 10"]
tickLabels = [0, 10, 20, 30, 40, 50]
graphTitle = 'Title Placeholder'
xAxisTitle, yAxisTitle = "x_title", "y_title"

Gridx, Gridy = 100, 60
labelSize, tickSize, titleSize, axisTitleSize = 1.2, 1.2, 1.8, 1.5
barThick, nsubs = 3, 10
nxlabels, nylabels = len(tickLabels), len(barLabels) # Added this in since not all cases will have the same number of x and y labels
t = nxlabels + nylabels

#Element spacing
step = Gridy/(nylabels + 1)#For the bars
tickStep = Gridx/(nxlabels-1) #For x labelsn
barLocation = [-1*(Gridy/2)+(k*step) for k in range(1, nylabels+1)] #Actual y coords of bars

xLabelHeight = -1*(Gridy/2)-1.4
# labelLocx = -1*(Gridx/2 + 2.6)
labelLocx = -5

LABEL_TIME_LOC = (100,30,1)
CAM_LOC = (50, 0, 170)
GRID_LINE_HEIGHT = 26

# current_max = 20

SPEED_SORT = 20 # frames per position change

FONT = "/home/philippy/youtube/00_tarantino_fucks/assets/fonts/Frutiger.ttf"

# COLOR_PALETTE = ["#fd7f6f", "#7eb0d5", "#b2e061", "#bd7ebe", "#ffb55a", "#ffee65", "#beb9db", "#fdcce5", "#8bd3c7"]
COLOR_PALETTE = ["201e1c","6b6763","56666e","484b6e","8e5798","8c5981","181f27","1c2929","6a3c4b","73609c"]


FPS = 192


DATA = [
    [1,2,5,10,20,2000,10_000],
    [5,6,4,100,1,15,2],
    [2,3,4,5,7,8000,10],
    [2,3,4,5,7,8000,10],
    [2,3,4,5,7,8000,5_000],
    [2,3,4,5,7,8000,10],
    [5,6,4,100,1,15,200],
    [5,10,20,100,1,15,2],
    [5,6,4,100,1,1_000,500],
    [5,6,4,50,1,1_500,4_000],
    ]



import bpy
import bmesh
from mathutils import Vector
from bmesh.types import BMVert

from blender.data import *
from blender.blenderpy.util import *
from blender.default_cfg import MODE_ANIMATION
from blender.blenderpy.graph import *

limits = []
x_pos = []
FR, MAX_FR, all_frames, all_frames_max, SUM, DATA_MAX, DATA_MIN = get_constants(DATA)

def handler_bar(scene):
    """updates size of bars"""
    frame_cur = scene.frame_current
    objList = [obj for obj in bpy.data.objects if 'cube' in obj.name]

    for i, obj in enumerate(objList[:len(DATA)]):
        hooks = [h for h in obj.children if "Hook" in h.name]
        for j, hook in enumerate(hooks):
            dist = all_frames[i][frame_cur] / all_frames_max[frame_cur] * 100
            pos_init = Vector((0, hook.location[1], hook.location[2]))
            hook.location = pos_init + Vector((dist, 0, 0))


def handler_timeframe(scene):
    """updates time object at keyframe"""
    text = bpy.data.objects['label_time']
    frame = scene.frame_current
    for i, val in enumerate(np.arange(0, MAX_FR, 48)):
        if frame >= val:
            text.data.body = f"timeframe {i}"
            # text.location = Vector((100,30,1))
            text.location = Vector(LABEL_TIME_LOC)


def handler_grid(scene):
    if MODE_ANIMATION == "static":
        frame = bpy.context.scene.frame_current
        labels = [obj for obj in bpy.data.objects if 'value_label' in obj.name]
        frame = bpy.context.scene.frame_current

        for i, cube in enumerate([c for c in bpy.data.objects if "cube" in c.name]):
            x_pos.append(dict())
            hooks = [h.location[0] for h in cube.children if "Hook" in h.name]
            print(hooks)
            x_pos[i][frame] = max(hooks)
        for i, label in enumerate(labels):
            if frame in x_pos[i]:
                value = "{:.2f}".format(x_pos[i][frame] * DATA_MAX / 100)
                label.data.body = value
                # print(value)

    if MODE_ANIMATION == "max":
        frame = bpy.context.scene.frame_current
        if frame <= MAX_FR:
            max_value = all_frames_max[frame]
        upper_limit = cceil(max_value)
        # lower_limit = upper_limit/10

        cur_limits = np.arange(0, upper_limit, upper_limit/4)
        cur_limits[0] = upper_limit
        for limit in cur_limits:
            if limit not in limits:
                limits.append(limit)

        for upper_limit in limits:
            name = f"upper_limit_{upper_limit}"
            if f"line_{name}" not in bpy.data.objects:
                add_line(name)

            for name in [line.name for line in bpy.data.objects if "line" in line.name]:
                value = float(name.split("_")[-1])
                line = bpy.data.objects[f"{name}"]
                line_label = bpy.data.objects[f"label_{name.split('line_')[-1]}"]

                line.location = Vector((value/max_value*100,0,1))
                line_label.location = Vector((line.location[0] - line_label.dimensions[0]/2,26.5,0))
                line_label.data.body = f"{value:g}"

                grid_material = bpy.data.materials.get(f"material_{name.split('line_')[-1]}")
                grid_material.use_nodes = True
                node_tree = grid_material.node_tree
                nodes = node_tree.nodes
                bsdf = nodes.get("Principled BSDF")

                if value/max_value <= 0.15:
                    bsdf.inputs['Alpha'].default_value = min(1, max(bsdf.inputs['Alpha'].default_value - 0.005, 0))
                elif value/max_value > 0.99:
                    bsdf.inputs['Alpha'].default_value = min(1, max(bsdf.inputs['Alpha'].default_value - 0.005, 0))
                else:
                    bsdf.inputs['Alpha'].default_value = min(1, max(bsdf.inputs['Alpha'].default_value + 0.0025, 0))


def handler_value_label(scene):
    frame = bpy.context.scene.frame_current
    for i, _ in enumerate(DATA):
        label = bpy.data.objects[f"value_label_{i}"]
        if frame <=MAX_FR:
            label.data.body = "{:.2f}".format(all_frames[i][frame])


def handler_pie(scene):
    bpy.ops.object.select_all(action='DESELECT')

    # Select the object
    if bpy.data.objects.get("pie_chart") is not None:
        bpy.data.objects['pie_chart'].select_set(True)

    bpy.ops.object.delete()

    pie_chart()

    # Select the object
    if bpy.data.objects.get("pie_chart") is not None:
        bpy.data.objects['pie_chart'].select_set(True)
        obj = bpy.data.objects['pie_chart']

    obj = bpy.data.objects['pie_chart']
    frame = bpy.context.scene.frame_current
    # all_frames = get_all_frames()
    # total = get_sum()

    per100 = [floor(all_frames[i][frame]/SUM[frame] * 100) for i,_ in enumerate(DATA)]
    if sum(per100) < 100:
        per100[per100.index(max(per100))] += 100 - sum(per100)

    EPSILON = 1.0e-5
    up_vector = Vector((0,0,1))

    bm = bmesh.new()
    bm.from_mesh(obj.data, face_normals=True)

    faces=[f for f in bm.faces if (f.normal-up_vector).length < EPSILON]

    bar2face = ([], [], [])
    for i, face in enumerate(faces):
        if i<=per100[0]:
            obj.data.polygons[face.index].material_index = 1
            bar2face[0].append(face.index)
        elif per100[0] < i < per100[0] + per100[1]:
            obj.data.polygons[face.index].material_index = 2
            bar2face[1].append(face.index)
        else:
            obj.data.polygons[face.index].material_index = 3
            bar2face[2].append(face.index)


    id_first = per100.index(sorted(per100)[-1])
    id_second = per100.index(sorted(per100)[-2])

    bm = bmesh.new()
    bm.from_mesh(obj.data, face_normals=True)
    faces = [f for f in bm.faces if f.index in bar2face[id_first]]
    extruded = bmesh.ops.extrude_face_region(bm, geom=faces)

    # Move extruded geometry
    translate_verts = [v for v in extruded['geom'] if isinstance(v, BMVert)]
    vec = up_vector * 0.5
    bmesh.ops.translate(bm, vec=vec, verts=translate_verts)

    # Second
    bm.faces.ensure_lookup_table()
    faces = [f for f in bm.faces if (f.normal-up_vector).length < EPSILON]
    faces = [f for f in bm.faces if f.index in bar2face[id_second]]
    extruded = bmesh.ops.extrude_face_region(bm, geom=faces)

    # Move extruded geometry
    translate_verts = [v for v in extruded['geom'] if isinstance(v, BMVert)]
    vec = up_vector * 0.25
    bmesh.ops.translate(bm, vec=vec, verts=translate_verts)

    # Update mesh and free Bmesh
    bm.normal_update()
    bm.to_mesh(obj.data)
    bm.free()



#Pie Graph Experiment Final Script

#import modules
import bpy, random

#Define a function for cleaning up text objects
def txtCleanup(title):
    bpy.ops.object.text_add(enter_editmode=True) #Generalise this; Done
    [bpy.ops.font.delete(type='PREVIOUS_OR_SELECTION') for j in range(4)]
    [bpy.ops.font.text_insert(text=char) for char in title]
    bpy.ops.object.editmode_toggle()
    bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')

#Define data and generate slice labels
data = [0, 0.2, 0.45, 0.13, 0.17, 0.05] ### First value in slice needs to be 0
pieLabels = [str(data[i]*100)+'%' for i in range(len(data))]

#Calculate slice positions and rotations
slice = [data[i]*2*3.14159 for i in range(len(data))] #pi = 3.14159
rot = [-1*(sum(slice[0:i+1])) for i in range(len(slice))]

#Define some cosmetic variables
pieHeight, titleY = 0.5, 2.5
labelZ = pieHeight + 0.05
labelSize, titleSize = 0.25, 1
graphTitle = "Pie Graph"

#Create the overhead title of the pie chart
bpy.context.scene.cursor.location = (0, 0, 0)
txtCleanup(graphTitle)
bpy.data.objects['Text'].location = (0, titleY, 0.2)
bpy.ops.transform.resize(value=(titleSize, titleSize, titleSize))

#Generate slices for pie graph and their labels
for i in range(len(data)-1):
    ### Slices
    bpy.context.scene.cursor.location = (0, 0, 0)
    bpy.ops.mesh.primitive_plane_add(enter_editmode=False, location=(0, 0, 0)) #New plane
    bpy.ops.transform.resize(value=(pieHeight, 1, 1)) #Scale in x
    bpy.ops.transform.translate(value=(0, 1, 0)) #move up
    bpy.ops.transform.rotate(value=1.5708, orient_axis='Y', orient_type='GLOBAL')
    bpy.ops.object.origin_set(type='ORIGIN_CURSOR', center='MEDIAN') #set origin
    bpy.context.object.rotation_euler[2] = rot[i] #rotate by the previous calculated angle
    bpy.ops.object.editmode_toggle() #Extrude face
    bpy.ops.mesh.extrude_region_move(MESH_OT_extrude_region={"use_normal_flip":False, "mirror":False}, TRANSFORM_OT_translate={"value":(0, 0, 0)})
    bpy.ops.mesh.spin(steps = 10, angle = slice[i+1], center=(0, 0, 0), axis=(0, 0, 1)) #Spin out

    ### Labels
    bpy.ops.object.editmode_toggle()
    bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='MEDIAN')
    bpy.context.scene.cursor.location = bpy.context.object.location
    txtCleanup(str(pieLabels[i+1]))
    bpy.context.object.location = bpy.context.scene.cursor.location
    bpy.context.object.location[2] = labelZ
    bpy.ops.transform.resize(value=(labelSize, labelSize, labelSize))

#Set cursor back to the origin
bpy.context.scene.cursor.location = (0, 0, 0)

#Add material to text
labelMat = bpy.data.materials.new(name='LabelMaterial')
labelMat.diffuse_color=(0, 0, 0, 1)
[bpy.data.objects[obj.name].data.materials.append(labelMat) for obj in bpy.data.objects if "Text" in obj.name]

#Add materials to bars
for obj in bpy.data.objects:
    if 'Plane' in obj.name:
        currentMat = bpy.data.materials.new(name='sliceMat' + obj.name)
        currentMat.diffuse_color=(random.random(), random.random(), random.random(), 1)
        bpy.data.objects[obj.name].data.materials.append(currentMat)



import bpy
import blender.blenderpy
from blender.blenderpy import Graph

def create(cfg):
    """creates graph based on cfg"""
    graph = Graph(cfg)
    blender.blenderpy.util.delete_all()
    graph.camera()
    graph.light()
    graph.background_plane()

    graph.bars()
    graph.title()
    graph.title_axis()
    graph.materials()
    graph.hooks()
    graph.label_value()
    graph.label_axis(mode="max")
    graph.label_timeframe()
    graph.grid(mode="max")

    #group graph elements
    graph.group_graph()
    graph.group_bars()
    graph.group_value_labels()

    #fixed locations
    graph.loc_fix_labels()

    #add callbacks
    bpy.app.handlers.frame_change_post.extend([graph.handler_bar, graph.handler_timeframe, graph.handler_grid, graph.handler_value_label])
    # bpy.app.handlers.frame_change_post.extend([handler_pie])

    #keyframe inserts
    graph.animate_sorting()
    # graph.move_camera()

    #setup render
    blender.blenderpy.util.set_render_settings()



import numpy as np

from blender.default_cfg import DATA, FPS

def get_all_frames(data):
    all_frames = []
    for j, bar in enumerate(data):
        all_frames.append([])
        for i, value in enumerate(bar[:-1]):
            next = bar[i+1]
            if value >= next:
                per_frame = np.arange(next, value, (value-next)/FPS)[::-1]
            else:
                per_frame = np.arange(value, next, (next-value)/FPS)
            if not per_frame.any():
                per_frame = [value] * FPS
            all_frames[j].extend(per_frame)
        all_frames[j].extend([bar[-1]] * FPS)
    return all_frames

def get_all_frames_max(data):
    all_frames_max = []
    all_frames = get_all_frames(data)
    for frame, value in enumerate(all_frames[0]):
        all_frames_max.append(max([bar[frame] for bar in all_frames]))
    return all_frames_max

def get_all_frames_sum(data):
    all_frames = get_all_frames(data)
    res = []
    for frame in range(0, get_LAST_FR(DATA)):
        total = sum([all_frames[i][frame] for i, _ in enumerate(data)])
        res.append(total)
    return res

def get_LAST_FR(data):
    LAST_FR = FPS*len(data[0]) #2 secs(48 frames) worth of animation for each datapoint
    return LAST_FR

def get_FR(data):
    FR = tuple(range(0, get_LAST_FR(DATA)+1, FPS))
    return FR

def get_data(data=None):
    if data:
        return data
    return DATA

def get_DATA_MAX(data):
    DATA_MAX = max([point for bar in DATA for point in bar])
    return DATA_MAX

def get_DATA_MIN(data):
    DATA_MIN = min([point for bar in DATA for point in bar])
    return DATA_MIN

def get_constants(data):
    FR = get_FR(data)
    LAST_FR = get_LAST_FR(data)
    all_frames = get_all_frames(data)
    all_frames_max = get_all_frames_max(data)
    SUM_FR = get_all_frames_sum(data)
    MAX_FR = get_DATA_MAX(data)
    MIN_FR = get_DATA_MIN(data)
    return FR, LAST_FR, all_frames, all_frames_max, SUM_FR, MAX_FR, MIN_FR



import pathlib
from setuptools import setup, find_packages

VERSION = '0.0.2'
DESCRIPTION = 'Data Driven Content Creation in Blender'
LONG_DESCRIPTION = ''

with open('requirements.txt') as f:
    required = f.read().splitlines()

# Setting up
setup(
        name="fsm-blender",
        version=VERSION,
        author="Philip Paprotny",
        author_email="<philip.paprotny@gmail.com>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        # packages=["nlpp"],
        include_package_data=True,
        install_requires=required,
        keywords=['python', 'blender', 'dataviz'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.9",
        ]
)



import pytest
import bpy
from bpy import context

import blender
from blender.default_cfg import *
from blender.graph import *
from blender.util import *

TEST_DATA = [
    [1.34,2,5,10,20,2000,20_000],
    [5,6,4,100,1,15,2],
    [2,3,4,5,7,8000,10]
    ]

# DATA = TEST_DATA

@pytest.fixture
def data():
    return TEST_DATA

@pytest.fixture(scope="session", autouse=True)
def setup_cfg(request):
    # prepare something ahead of all tests
    blender.default_cfg.DATA = TEST_DATA


def test_cfg():
    assert TEST_DATA
    assert blender.default_cfg.DATA == TEST_DATA
    assert blender.default_cfg.COLOR_PALETTE[-1] == "73609c"
    assert blender.default_cfg.DATA[0][0] == 1.34
    assert blender.default_cfg.DATA[0][0] != 1

def test_util_add_and_delete():
    bpy.ops.mesh.primitive_cube_add(size = 1)
    context.active_object.name = f"cube_test"
    assert len(bpy.data.objects) == 1
    delete_all()
    assert len(bpy.data.objects) == 0

def test_util_error():
    with pytest.raises(ZeroDivisionError) as excinfo:
        force_div_by_zero_error()
    # assert str(excinfo.value) == 'division by zero'

def test_util_value_error_match():
    # with pytest.raises(ValueError, match=r'must be \d+$'):
    with pytest.raises(ValueError, match=r'math domain'):
        force_value_error()

def test_util_value_error_match_custom_message():
    with pytest.raises(CustomError) as excinfo:
        force_value_error_custom(-2)
        assert excinfo.value.message == 'everything is broken'

def test_data_fix(data):
    assert data[0][0] == 1.34



import bpy
import bmesh
import math


# ------------------------------------------------------------------------------
# Utility Functions

def set_smooth(obj):
    """ Enable smooth shading on an mesh object """

    for face in obj.data.polygons:
        face.use_smooth = True


def object_from_data(data, name, scene, select=True):
    """ Create a mesh object and link it to a scene """

    mesh = bpy.data.meshes.new(name)
    mesh.from_pydata(data['verts'], data['edges'], data['faces'])

    obj = bpy.data.objects.new(name, mesh)
    scene.collection.objects.link(obj)

    bpy.context.view_layer.objects.active = obj
    obj.select_set(True)

    mesh.update(calc_edges=True)
    mesh.validate(verbose=True)

    return obj


def recalculate_normals(mesh):
    """ Make normals consistent for mesh """

    bm = bmesh.new()
    bm.from_mesh(mesh)

    bmesh.ops.recalc_face_normals(bm, faces=bm.faces)

    bm.to_mesh(mesh)
    bm.free()


# ------------------------------------------------------------------------------
# Geometry functions

def vertex_circle(segments, z):
    """ Return a ring of vertices """
    verts = []

    for i in range(segments):
        angle = (math.pi*2) * i / segments
        verts.append((math.cos(angle), math.sin(angle), z))

    return verts


def face(segments, i, row):
    """ Return a face on a cylinder """

    if i == segments - 1:
        ring_start = segments * row
        base = segments * (row + 1)

        return (base - 1, ring_start, base, (base + segments) - 1)

    else:
        base = (segments * row) + i
        return (base, base + 1, base + segments + 1, base + segments)


def bottom_cap(verts, faces, segments, cap='NGON'):
    """ Build bottom caps as triangle fans """

    if cap == 'TRI':
        verts.append((0, 0, 0))
        center_vert = len(verts) - 1

        [faces.append((i, i+1, center_vert)) for i in range(segments - 1)]
        faces.append((segments - 1, 0, center_vert))

    elif cap == 'NGON':
        faces.append([i for i in range(segments)])

    else:
        print('[!] Passed wrong type to bottom cap')


def top_cap(verts, faces, segments, rows, cap='NGON'):
    """ Build top caps as triangle fans """

    if cap == 'TRI':
        verts.append((0, 0, rows - 1))
        center_vert = len(verts) - 1
        base = segments * (rows - 1)

        [faces.append((base+i, base+i+1, center_vert))
                       for i in range(segments - 1)]

        faces.append((segments * rows - 1, base, center_vert))

    elif cap == 'NGON':
        base = (rows - 1) * segments
        faces.append([i + base for i in range(segments)])

    else:
        print('[!] Passed wrong type to top cap')


# ------------------------------------------------------------------------------
# Main Functions

def make_circle(name, segments=32, fill=None):
    """ Make a circle """

    data = {
            'verts': vertex_circle(segments, 0),
            'edges': [],
            'faces': [],
           }

    if fill:
        bottom_cap(data['verts'], data['faces'], segments, fill)
    else:
        data['edges'] = [(i, i+1) for i in range(segments)]
        data['edges'].append((segments - 1, 0))

    scene = bpy.context.scene
    return object_from_data(data, name, scene)


def make_cylinder(name, segments=64, rows=4, cap=None):
    """ Make a cylinder """

    data = { 'verts': [], 'edges': [], 'faces': [] }

    for z in range(rows):
        data['verts'].extend(vertex_circle(segments, z))

    for i in range(segments):
        for row in range(0, rows - 1):
            data['faces'].append(face(segments, i, row))

    if cap:
        bottom_cap(data['verts'], data['faces'], segments, cap)
        top_cap(data['verts'], data['faces'], segments, rows, cap)


    scene = bpy.context.scene
    obj = object_from_data(data, name, scene)
    recalculate_normals(obj.data)
    set_smooth(obj)

    bevel = obj.modifiers.new('Bevel', 'BEVEL')
    bevel.limit_method = 'ANGLE'

    obj.modifiers.new('Edge Split', 'EDGE_SPLIT')

    return obj


# ------------------------------------------------------------------------------
# Main Code

#make_circle('Circle', 64)
# make_cylinder('Cylinder', 100, 2, 'TRI')
# bpy.ops.object.editmode_toggle()
# bpy.ops.mesh.inset(use_boundary=False, use_even_offset=False, use_relative_offset=False, use_edge_rail=False, thickness=0.307055, depth=0, use_outset=False, use_select_inset=False, use_individual=False, use_interpolate=True)
# bpy.ops.object.editmode_toggle()

#bpy.ops.mesh.bridge_edge_loops()
